-- Enables creation of uuids
CREATE EXTENSION IF NOT EXISTS "uuid-ossp";

CREATE SCHEMA core;

CREATE TYPE core.party_kind_t AS ENUM ('PERSON', 'ORGANISATION');
CREATE TYPE core.doc_approval_status_t AS ENUM (
  'DRAFT',
  'UNDER_REVIEW',
  'APPROVED'
);

--
-- Tables
--

CREATE TABLE core.object (
  object_uuid uuid DEFAULT uuid_generate_v4() PRIMARY KEY,
  schema      name,
  name        name,
  created     timestamptz DEFAULT CURRENT_TIMESTAMP
);

CREATE TABLE core.search (
  object_uuid uuid REFERENCES core.object (object_uuid) ON DELETE CASCADE,
  tsv         tsvector,

  PRIMARY KEY(object_uuid)
);

CREATE TABLE core.country (
  alpha_2 text PRIMARY KEY,
  pretty  text
);

/**
 * This table is based on ISO 3166-2: Country subdivison code
 * In other words, states/provinces
 */
CREATE TABLE core.subdivision (
  alpha_2 text REFERENCES core.country (alpha_2),
  code    text,
  pretty  text,
  PRIMARY KEY (alpha_2, code)
);

CREATE TABLE core.language (
  alpha_2      text,
  alpha_3      text PRIMARY KEY,
  english_name text
);

CREATE TABLE core.address (
  address_uuid uuid DEFAULT uuid_generate_v4() PRIMARY KEY,
  addr1         text,
  addr2         text,
  locality_name text,
  subdivision   text, -- state/province
  country       text,
  code          text,
  type          smallint, -- 0: residential, 1: business
  created       timestamptz DEFAULT CURRENT_TIMESTAMP,
  FOREIGN KEY (subdivision, country) REFERENCES core.subdivision (code, alpha_2)
);

CREATE TABLE core.full_address (
  address_uuid  uuid DEFAULT uuid_generate_v4() PRIMARY KEY,
  lot_number    text,
  road_number1  text,
  road_number2  text,
  road_name     text,
  road_type     text,
  road_suffix   text,
  locality_name text,
  subdivision   text, -- state/province
  country       text,
  code          text,
  type          smallint, -- 0: residential, 1: business
  created timestamptz DEFAULT CURRENT_TIMESTAMP,
  FOREIGN KEY (subdivision, country) REFERENCES core.subdivision (code, alpha_2)
);

CREATE TABLE core.party (
  party_uuid           uuid REFERENCES core.object (object_uuid) ON DELETE CASCADE,
  kind                 core.party_kind_t,
  name                 text NOT NULL,
  email                text UNIQUE,
  uri                  text UNIQUE, -- Email for a person, URL for an organization
  data                 jsonb,
  address_uuid         uuid REFERENCES core.address (address_uuid) ON DELETE SET NULL,
  billing_address_uuid uuid REFERENCES core.address (address_uuid) ON DELETE SET NULL,
  language_code        text REFERENCES core.language (alpha_3) ON DELETE SET NULL,
  timezone_abbrev      text,
  created              timestamptz DEFAULT CURRENT_TIMESTAMP,
  modified             timestamptz,

  PRIMARY KEY (party_uuid),
  CHECK (email IS NOT NULL OR uri IS NOT NULL)
);

CREATE TABLE core.person (
  party_uuid           uuid REFERENCES core.party (party_uuid) ON DELETE CASCADE,
  mobile               text,
  phone                text,
  PRIMARY KEY (party_uuid)
);

CREATE TABLE core.organisation (
  party_uuid           uuid REFERENCES core.party (party_uuid) ON DELETE CASCADE,
  trading_name         text,
  url                  text,
  industry_code        text,
  PRIMARY KEY (party_uuid)
);

CREATE TABLE core.contact (
  contact_uuid  uuid REFERENCES core.party (party_uuid) ON DELETE CASCADE,
  PRIMARY KEY (contact_uuid)
);

CREATE TABLE core.role (
  role_id serial PRIMARY KEY,
  name    text
);

CREATE TABLE core.relationship (
  relationship_id serial PRIMARY KEY,
  name            text,
  description     text,
  -- The parties involved in the relationship, the first role should be the
  -- more superior
  first_role_id   integer REFERENCES core.role (role_id) ON DELETE CASCADE,
  second_role_id  integer REFERENCES core.role (role_id) ON DELETE CASCADE,
  created         timestamptz DEFAULT CURRENT_TIMESTAMP
);

CREATE TABLE core.party_relationship (
  relationship_id   integer REFERENCES core.relationship (relationship_id) ON DELETE CASCADE,
  first_party_uuid  uuid REFERENCES core.party (party_uuid) ON DELETE CASCADE,
  second_party_uuid uuid REFERENCES core.party (party_uuid) ON DELETE CASCADE,
  ended             timestamptz,
  created           timestamptz DEFAULT CURRENT_TIMESTAMP
);

-- A place to put DB related settings
CREATE TABLE core.setting (
  name  text,
  value text
);

CREATE TABLE core.document_status (
  status_id serial PRIMARY KEY,
  name      text
);

CREATE TABLE core.document (
  document_uuid   uuid DEFAULT uuid_generate_v4() PRIMARY KEY,
  document_num    bigserial,
  name            name, -- Name of linking table
  schema          name, -- Schema of linking table
  status_id       integer REFERENCES core.document_status (status_id) NOT NULL,
  data            jsonb,
  created         timestamptz DEFAULT CURRENT_TIMESTAMP
);

CREATE TABLE core.note (
  note_uuid  uuid DEFAULT uuid_generate_v4() PRIMARY KEY,
  body       text,
  public     boolean DEFAULT FALSE,
  created    timestamptz DEFAULT CURRENT_TIMESTAMP,
  created_by uuid REFERENCES core.party (party_uuid)
);

CREATE TABLE core.tag (
  tag_uuid uuid DEFAULT uuid_generate_v4() PRIMARY KEY,
  name     text UNIQUE
);

--
-- Views
--

CREATE OR REPLACE VIEW core.person_v AS
  SELECT
    p.*,
    prsn.mobile,
    prsn.phone
  FROM core.party p
  LEFT JOIN core.person prsn
    USING (party_uuid);

CREATE OR REPLACE VIEW core.organisation_v AS
  SELECT
    p.*,
    org.trading_name,
    org.url,
    org.industry_code
  FROM core.party p
  LEFT JOIN core.organisation org
    USING (party_uuid);

CREATE OR REPLACE VIEW core.address_v AS
  SELECT
    ad.address_uuid,
    ad.addr1,
    ad.addr2,
    ad.locality_name,
    s.pretty AS subdivision,
    c.pretty AS country,
    ad.code,
    ad.type
  FROM core.address ad
  LEFT JOIN core.country c
    ON c.alpha_2 = ad.country
  LEFT JOIN core.subdivision s
    ON s.code = ad.subdivision;

CREATE OR REPLACE VIEW core.document_v AS
  SELECT
    d.*,
    s.name AS document_status
  FROM core.document d
  LEFT JOIN core.document_status s
    USING (status_id);

CREATE OR REPLACE VIEW core.relationship_v AS
  SELECT
    r.relationship_id,
    r.name,
    r.description,
    rlf.name AS first_role_name,
    rls.name AS second_role_name,
    r.created
  FROM core.relationship r
  LEFT JOIN core.role rlf
    ON rlf.role_id = r.first_role_id
  LEFT JOIN core.role rls
    ON rls.role_id = r.second_role_id;

--
-- Triggers
--

CREATE OR REPLACE FUNCTION core.insert_party_trigger () RETURNS TRIGGER AS
$$
BEGIN
  INSERT INTO core.object (
    name,
    schema
  ) VALUES (
    TG_TABLE_NAME,
    TG_TABLE_SCHEMA
  )
  RETURNING object_uuid INTO NEW.party_uuid;

  RETURN NEW;
END
$$
LANGUAGE 'plpgsql';

CREATE TRIGGER core_insert_party BEFORE INSERT ON core.party
FOR EACH ROW EXECUTE PROCEDURE core.insert_party_trigger();
