-- required: core
-- required: scm

CREATE SCHEMA prj;

CREATE TYPE prj.job_status_t AS ENUM ('WIP', 'COMPLETE');

CREATE TABLE prj.state (
  state_uuid uuid DEFAULT uuid_generate_v4() PRIMARY KEY,
  name       text
);

CREATE TABLE prj.project (
  project_uuid uuid DEFAULT uuid_generate_v4() PRIMARY KEY,
  address_uuid uuid REFERENCES core.full_address (address_uuid) ON DELETE RESTRICT,
  owner_uuid   uuid REFERENCES core.party (party_uuid) ON DELETE RESTRICT,
  name         text NOT NULL
);

CREATE TABLE prj.role (
  role_uuid uuid DEFAULT uuid_generate_v4() PRIMARY KEY,
  name      text
);

CREATE TABLE prj.project_role (
  project_uuid uuid REFERENCES prj.project (project_uuid) ON DELETE CASCADE,
  party_uuid   uuid REFERENCES core.party (party_uuid) ON DELETE CASCADE
);

CREATE TABLE prj.object (
  object_uuid  uuid REFERENCES core.object (object_uuid) ON DELETE CASCADE,
  project_uuid uuid REFERENCES prj.project (project_uuid),
  PRIMARY KEY (object_uuid)
);

CREATE TABLE prj.job (
  job_uuid        uuid REFERENCES prj.object (object_uuid) ON DELETE CASCADE,
  dependant_uuid  uuid REFERENCES prj.job (job_uuid) ON DELETE CASCADE,
  state           uuid REFERENCES prj.state (state_uuid) ON DELETE RESTRICT,
  id              text,
  name            text NOT NULL,
  short_desc      text,
  description     text,
  assignee        uuid REFERENCES core.party (party_uuid) ON DELETE SET NULL,
  sequence_num    integer NOT NULL,
  duration        interval,
  lag             interval,
  lead            interval,
  finished_at     timestamptz,
  created_by      uuid REFERENCES core.person (party_uuid) ON DELETE SET NULL,
  created         timestamptz DEFAULT CURRENT_TIMESTAMP,
  modified        timestamptz DEFAULT CURRENT_TIMESTAMP,
  PRIMARY KEY (job_uuid)
);

CREATE TABLE prj.deliverable (
  deliverable_uuid uuid REFERENCES prj.object (object_uuid) ON DELETE CASCADE,
  job_uuid         uuid REFERENCES prj.job (job_uuid) ON DELETE CASCADE,
  item_uuid        uuid REFERENCES scm.item (item_uuid) ON DELETE RESTRICT,
  sequence_num     integer,
  data             jsonb,
  created          timestamptz DEFAULT CURRENT_TIMESTAMP,
  PRIMARY KEY (deliverable_uuid)
);

CREATE TABLE prj.job_invoice (
  job_uuid     uuid REFERENCES prj.job (job_uuid) ON DELETE RESTRICT,
  invoice_uuid uuid REFERENCES ar.invoice (invoice_uuid) ON DELETE RESTRICT,
  PRIMARY KEY (job_uuid, invoice_uuid)
);

--
-- Associative Tables
--

CREATE TABLE prj.project_job (
  project_uuid uuid REFERENCES prj.project (project_uuid) ON DELETE CASCADE,
  job_uuid     uuid REFERENCES prj.job (job_uuid) ON DELETE CASCADE
);

--
-- Triggers
--

CREATE OR REPLACE FUNCTION prj.insert_project_trigger() RETURNS trigger AS
$$
DECLARE
  new_job_uuid uuid;
BEGIN
  WITH new_job AS (
    INSERT INTO prj.job (name) VALUES (NEW.name)
    RETURNING *
  )
  INSERT INTO prj.project_job (
    project_uuid,
    job_uuid
  ) VALUES (
    NEW.project_uuid,
    (SELECT job_uuid FROM new_job)
  )
  RETURNING job_uuid INTO new_job_uuid;

  UPDATE prj.object o
  SET project_uuid = NEW.project_uuid
  WHERE o.object_uuid = new_job_uuid;

  RETURN new;
END
$$ LANGUAGE 'plpgsql';

CREATE TRIGGER prj_project_insert AFTER INSERT ON prj.project
FOR EACH ROW EXECUTE PROCEDURE prj.insert_project_trigger();

/**
 * Create objects when inserting into job.
 * When a new job is inserted, the project to which it belongs is
 * determined by the job's dependant. The root job of a project has no dependant
 * so jobs that have the root as their dependant need to look up the project
 * from prj.project_job.
 */
CREATE OR REPLACE FUNCTION prj.insert_job_trigger() RETURNS trigger AS
$$
DECLARE
  new_object_uuid uuid;
BEGIN
  WITH core_object AS (
    INSERT INTO core.object (
      schema,
      name
    ) VALUES (
      TG_TABLE_SCHEMA,
      TG_TABLE_NAME
    )
    RETURNING *
  )
  INSERT INTO prj.object (
    object_uuid,
    project_uuid
  ) VALUES (
    (SELECT object_uuid FROM core_object),
    (
      SELECT
        o.project_uuid
      FROM prj.object o
      WHERE o.object_uuid = NEW.dependant_uuid
    )
  )
  RETURNING object_uuid INTO new_object_uuid;
  
  NEW.job_uuid = new_object_uuid;

  RETURN new;
END
$$ LANGUAGE 'plpgsql';

CREATE TRIGGER prj_job_insert BEFORE INSERT ON prj.job
FOR EACH ROW EXECUTE PROCEDURE prj.insert_job_trigger();

/**
 * Create new deliverable with object
 */
CREATE OR REPLACE FUNCTION prj.insert_deliverable_trigger() RETURNS trigger AS
$$
DECLARE
  new_object_uuid uuid;
BEGIN
  WITH core_object AS (
    INSERT INTO core.object (
      schema,
      name
    ) VALUES (
      TG_TABLE_SCHEMA,
      TG_TABLE_NAME
    )
    RETURNING *
  )
  INSERT INTO prj.object (
    object_uuid,
    project_uuid
  ) VALUES (
    (SELECT object_uuid FROM core_object),
    (
      SELECT
        o.project_uuid
      FROM prj.object o
      WHERE o.object_uuid = NEW.job_uuid
    )
  )
  RETURNING object_uuid INTO new_object_uuid;
  
  NEW.deliverable_uuid = new_object_uuid;

  RETURN new;
END
$$ LANGUAGE 'plpgsql';

CREATE TRIGGER prj_deliverable_insert BEFORE INSERT ON prj.deliverable
FOR EACH ROW EXECUTE PROCEDURE prj.insert_deliverable_trigger();

/**
 * Delete deliverable
 */
CREATE OR REPLACE FUNCTION prj.delete_deliverable_trigger() RETURNS trigger AS
$$
BEGIN
  DELETE FROM scm.item WHERE item_uuid = OLD.item_uuid;
  RETURN OLD;
END
$$ LANGUAGE 'plpgsql';

CREATE TRIGGER prj_deliverable_delete AFTER DELETE ON prj.deliverable
FOR EACH ROW EXECUTE PROCEDURE prj.delete_deliverable_trigger();

-- Create default states
INSERT INTO prj.state ( name ) VALUES
  ('UNSTARTED'),
  ('INPROGRESS'),
  ('FINISHED');
