/**
 * Generate the data for an invoice document
 * 
 * When a detail's quantity is NULL, Terbo assumes that it means 1.000
 */

CREATE OR REPLACE FUNCTION sales.invoice_document (json, OUT result json) AS
$$
BEGIN
  WITH invoice AS (
    SELECT
      i.*,
      d.*
    FROM ar.invoice i
    LEFT JOIN core.document d
      ON d.document_uuid = i.invoice_uuid
    WHERE i.invoice_uuid = ($1->>'invoice_uuid')::uuid
  ),
  details AS (
    SELECT
      id.*,
      li.line_item_uuid,
      li.order_uuid,
      li.product_uuid,
      li.code,
      li.name,
      li.short_desc,
      li.quantity,
      li.uom_name,
      li.uom_abbr,
      li.product_short_desc,
      li.price,
      (COALESCE(li.quantity, 1.000) * li.price)::numeric(10,2) AS total
    FROM ar.invoice_detail id
    LEFT JOIN sales.line_items(_line_item_uuid => id.detail_uuid) li
      ON li.line_item_uuid = id.detail_uuid
    WHERE id.invoice_uuid = ($1->>'invoice_uuid')::uuid
  ),
  customer AS (
    SELECT
      cv.customer_uuid,
      cv.name,
      cv.kind,
      ad.addr1,
      ad.addr2,
      ad.locality,
      ad.subdivision,
      ad.country,
      ad.code
    FROM sales.customer_v cv
    LEFT JOIN core.address_v ad
      ON ad.address_uuid = cv.address_uuid
    WHERE cv.customer_uuid = (SELECT payor_uuid FROM invoice)
  ), contact AS (
    SELECT
      p.name,
      pe.email
    FROM core.party p
    LEFT JOIN core.person pe
      USING (party_uuid)
    WHERE p.party_uuid = (SELECT contact_uuid FROM invoice)
  ),
  seller AS (
    SELECT
      p.party_uuid,
      p.name,
      p.kind,
      ad.addr1,
      ad.addr2,
      ad.locality,
      ad.subdivision,
      ad.country,
      ad.code
    FROM core.party p
    LEFT JOIN core.address_v ad
      ON ad.address_uuid = p.billing_address_uuid
    WHERE p.party_uuid = (SELECT seller_uuid FROM invoice)
  ),
  seller_contact AS (
    SELECT
      p.party_uuid,
      p.name,
      p.email
    FROM core.party p
    WHERE p.party_uuid = (SELECT seller_contact_uuid FROM invoice)
  )
  SELECT INTO result
    json_strip_nulls(to_json(r))
  FROM (
    SELECT
      i.*,
      to_char(i.due_date, core.setting('default_date_format')) AS due_date,
      (
        SELECT
          c
        FROM customer c
      ) AS payor,
      (
        SELECT
          c
        FROM contact c
      ) AS contact,
      (SELECT s FROM seller s) AS seller,
      (SELECT sc FROM seller_contact sc) AS seller_contact,
      (
        SELECT DISTINCT ON (d.order_uuid)
          array_agg(d.order_uuid)
        FROM details d
        GROUP BY d.order_uuid
      ) AS order_uuid,
      (
        SELECT
          array_agg(d)
        FROM (
          SELECT
            id.*
          FROM details id
        ) d
      ) AS details,
      (
        SELECT
          sum(d.total)::numeric(10,2)
        FROM details d
      ) AS subtotal,
      (
        SELECT
          (sum(d.total) * 1.1)::numeric(10,2) -- TODO do taxes properly
        FROM details d
      ) AS amount_payable,
      (
        SELECT
          ((sum(d.total) * 1.1) - sum(d.total))::numeric(10,2)
        FROM details d
      ) AS tax_payable
    FROM invoice i
  ) r;
END
$$
LANGUAGE 'plpgsql' SECURITY DEFINER;
