/**
 * Return the relevant price for a product.
 * Where a product has no associated price, attempt to generate one if the
 * product is an assembly or a composite.
 * Where a price is not found, the absence of a price_uuid indicates the price
 * is generated.
 * Optionally, a timestamp can be passed as the second argument to determine the
 * at a specific point in time.
 */

CREATE OR REPLACE FUNCTION sales.product_price (
  uuid,
  timestamptz DEFAULT CURRENT_TIMESTAMP,
  OUT result sales.product_price_v
) AS
$$
DECLARE
  cost   prd.cost;
  price  sales.price;
  profit numeric(10,2);
BEGIN
  SELECT
    *
  INTO cost
  FROM prd.cost($1);
  
  SELECT DISTINCT ON (pp.product_uuid)
    pr.*
  INTO price
  FROM sales.product_price pp
  INNER JOIN sales.price pr
    ON pr.price_uuid = pp.price_uuid
  WHERE pp.product_uuid = $1 AND
    pr.created < $2 AND
    (pr.end_at IS NULL OR pr.end_at < $2)
  ORDER BY pp.product_uuid, pr.created DESC;
  
  -- Get any named margins or markups. Since margin has precedence, get it
  -- before markup
  IF price.margin_id IS NOT NULL THEN
    SELECT
      mg.amount INTO price.margin
    FROM sales.margin mg
    WHERE mg.margin_id = price.margin_id;
  ELSIF price.markup_id IS NOT NULL THEN
    SELECT
      mk.amount INTO price.markup
    FROM sales.markup mk
    WHERE mk.markup_id = price.markup_id;
  END IF;
  
  IF price.amount_payable IS NOT NULL THEN
    result.amount_payable = price.amount_payable;
    result.price = (price.amount_payable * 0.909090909)::numeric(10,2);
  ELSIF price.price IS NOT NULL THEN
    result.price = price.price;
    
    IF cost.amount IS NOT NULL THEN
      profit = result.price - cost.amount;
      result.margin = (profit / result.price)::numeric(4,3);
    END IF;
  ELSIF cost.amount IS NOT NULL AND price.margin IS NOT NULL THEN
    result.price = (cost.amount / (1.000 - price.margin));
    result.margin = price.margin;
  ELSIF cost.amount IS NOT NULL THEN
    result.price = cost.amount;
  END IF;

  -- The default tax_excluded value is false
  result.tax_excluded = COALESCE(price.tax_excluded, FALSE);

  RETURN QUERY
  SELECT
    $1,
    result.price,
    result.margin;
END
$$
LANGUAGE 'plpgsql';

CREATE OR REPLACE FUNCTION sales.product_price (json, OUT result json)
AS
$$
BEGIN
  SELECT
    json_strip_nulls(to_json(r))
  INTO
    result
  FROM sales.product_price(($1->>'product_uuid')::uuid) r;
END
$$
LANGUAGE 'plpgsql' SECURITY DEFINER;
