-- required: prd

-- Enables creation of uuids
CREATE EXTENSION IF NOT EXISTS "uuid-ossp";

CREATE SCHEMA scm;

CREATE TYPE scm.item_kind_t AS ENUM ('ITEM', 'ASSEMBLY', 'PART', 'PRODUCT');
CREATE TYPE scm.item_t AS (
  item_uuid          uuid,
  product_uuid       uuid,
  kind               scm.item_kind_t,
  name               text,
  short_desc         text,
  data               jsonb,
  product_name       text,
  product_short_desc text,
  product_code       text,
  sku                text,
  route_uuid         uuid,
  weight             numeric(10,3),
  created            timestamptz,
  end_at             timestamptz,
  modified           timestamptz
);

CREATE TABLE scm.item (
  item_uuid      uuid DEFAULT uuid_generate_v4() PRIMARY KEY,
  product_uuid   uuid REFERENCES prd.product (product_uuid) ON DELETE RESTRICT,
  kind           scm.item_kind_t,
  name           text,
  short_desc     text,
  data           jsonb,
  weight         numeric(10,3),
  location       uuid,
  next_location  uuid,
  created        timestamptz DEFAULT CURRENT_TIMESTAMP,
  end_at         timestamptz,
  modified       timestamptz DEFAULT CURRENT_TIMESTAMP
);

CREATE TABLE scm.attribute (
  attribute_uuid uuid DEFAULT uuid_generate_v4() PRIMARY KEY,
  item_uuid      uuid REFERENCES scm.item (item_uuid) ON DELETE CASCADE,
  name           text,
  value          text,
  created        timestamptz DEFAULT CURRENT_TIMESTAMP,
  modified       timestamptz
);

CREATE TABLE scm.part (
  part_uuid   uuid DEFAULT uuid_generate_v4() PRIMARY KEY,
  root_uuid   uuid REFERENCES scm.item (item_uuid) ON DELETE CASCADE,
  parent_uuid uuid REFERENCES scm.item (item_uuid) ON DELETE CASCADE,
  item_uuid   uuid REFERENCES scm.item (item_uuid) ON DELETE RESTRICT -- Can't delete an Item if it's part of another Item
);

CREATE TABLE scm.component (
  component_uuid uuid DEFAULT uuid_generate_v4() PRIMARY KEY,
  item_uuid      uuid REFERENCES scm.item (item_uuid) ON DELETE CASCADE,
  product_uuid   uuid REFERENCES prd.product (product_uuid) ON DELETE CASCADE,
  quantity       numeric(10,3),
  data           jsonb,
  line_position  integer
);

CREATE TABLE scm.bom (
  bom_uuid   uuid DEFAULT uuid_generate_v4() PRIMARY KEY,
  created    timestamptz DEFAULT CURRENT_TIMESTAMP
);

CREATE TABLE scm.bom_line (
  bom_uuid     uuid REFERENCES scm.bom (bom_uuid) ON DELETE CASCADE,
  product_uuid uuid REFERENCES prd.product (product_uuid) ON DELETE SET NULL,
  quantity     numeric(10,2),
  modified     timestamptz DEFAULT CURRENT_TIMESTAMP,
  created      timestamptz DEFAULT CURRENT_TIMESTAMP
);

CREATE TABLE scm.boq (
  boq_uuid uuid DEFAULT uuid_generate_v4() PRIMARY KEY,
  created  timestamptz DEFAULT CURRENT_TIMESTAMP
);

CREATE TABLE scm.boq_line (
  boq_uuid     uuid DEFAULT uuid_generate_v4() PRIMARY KEY,
  product_uuid uuid REFERENCES prd.product (product_uuid) ON DELETE SET NULL,
  quantity     numeric(10,2),
  modified     timestamptz DEFAULT CURRENT_TIMESTAMP,
  created      timestamptz DEFAULT CURRENT_TIMESTAMP
);

CREATE VIEW scm.item_v AS
  SELECT
    i.*,
    p.name AS product_name
  FROM scm.item i
  LEFT JOIN prd.product p
    USING (product_uuid);

/*
 * Triggers
 */

CREATE OR REPLACE FUNCTION scm.delete_part_trg () RETURNS TRIGGER AS
$$
BEGIN
  DELETE FROM scm.item WHERE item_uuid = OLD.item_uuid;
  RETURN OLD;
END
$$
LANGUAGE 'plpgsql';

CREATE TRIGGER scm_delete_part_trg AFTER DELETE ON scm.part
FOR EACH ROW EXECUTE PROCEDURE scm.delete_part_trg();
